#!/usr/bin/env python

#
# The ESig Python package
# Basic package functions
#

import functools
import os

import numpy


ESIG_PACKAGE_ROOT = os.path.dirname(os.path.abspath(__file__))

# Python 3.8+ doesn't use PATH to locate DLLs; need to use add_dll_directory instead.
# Not sure if this is the right place to do this, though; 3.8 Porting Guide says to do
# this "while loading your library" (https://docs.python.org/3/whatsnew/3.8.html#bpo-36085-whatsnew).
# Note that add_dll_directory doesn't exist on non-Windows platforms or before Python 3.8.
import sys
try:
    from os.path import expanduser
    recombine_dll_dir = ESIG_PACKAGE_ROOT
    os.add_dll_directory(recombine_dll_dir)
except AttributeError:
    pass

from esig.backends import get_backend, set_backend, list_backends
from esig.tosig import recombine, NO_RECOMBINE


__all__ = [
    "get_version",
    "is_library_loaded",
    "get_library_load_error",
    "stream2sig",
    "stream2logsig",
    "logsigdim",
    "sigdim",
    "sigkeys",
    "logsigkeys",
    "recombine",
    "get_backend",
    "set_backend",
    "list_backends",
    "backends"
]


def get_version():
    """
    Returns the version number of the ESig package.
    The version is obtained from the VERSION file in the root of the package.

    Args:
        None

    Returns:
        string: The package version number. In format 'major.minor.release'.
    """
    version_filename = os.path.join(ESIG_PACKAGE_ROOT, 'VERSION')
    f = open(version_filename, 'r')
    version_string = f.read().strip().split(' ')
    f.close()

    return '.'.join(version_string)

def is_library_loaded():
    """
    Determines whether the tosig Python extension can be successfully loaded.
    If the library cannot be loaded successfully, debugging information can be obtained from get_library_load_error().

    Args:
        None
    Returns:
        boolean: True iif the library can be loaded successfully; False otherwise.
    """
    try:
        from esig import tosig
    except ImportError:
        return False

    return True


def get_library_load_error():
    """
    Returns a string containing the message of the exception raised when attempting to import tosig.
    If no exception is raised when attempting to import, None is returned.

    Args:
        None
    Returns:
        string: The message associated with the exception when attempting to import.
        None: If no exception is raised when importing, None is returned.
    """
    try:
        from esig import tosig
        return None
    except ImportError as e:
        return e.msg


def _verify_stream_arg(*types):
    """
    Helper decorator to provide type checking on the
    Numpy arrays
    """
    if types and callable(types[0]):
        fn = types[0]
        types = None
    else:
        fn = None

    types = types or (numpy.float32, numpy.float64)

    def decorator(func):

        @functools.wraps(func)
        def wrapper(stream, *args, **kwargs):
            as_array = numpy.array(stream)
            if not as_array.dtype in types:
                str_types = tuple(map(str, types))
                raise TypeError("Values must be of one of the following types {}".format(str_types))

            return func(as_array, *args, **kwargs)
        return wrapper

    if fn:
        return decorator(fn)
    return decorator


@_verify_stream_arg
def stream2sig(stream, depth):
    """
    Compute the signature of a stream
    """
    if depth <= 0:
        raise ValueError("Depth must be at least 1")
    elif depth == 1:
        return numpy.concatenate([[1.0], numpy.sum(numpy.diff(stream, axis=0), axis=0)])

    backend = get_backend()
    return backend.compute_signature(stream, depth)


@_verify_stream_arg
def stream2logsig(stream, depth):
    """
    Compute the log signature of a stream
    """
    if depth <= 0:
        raise ValueError("Depth must be at least 1")
    elif depth == 1:
        return numpy.sum(numpy.diff(stream, axis=0), axis=0)

    backend = get_backend()
    return backend.compute_log_signature(stream, depth)


def logsigdim(dimension, depth):
    """
    Get the number of elements in the log signature
    """
    if dimension == 0:
        raise ValueError("Dimension 0 is invalid")
    if depth == 1:
        return dimension
    return get_backend().log_sig_dim(dimension, depth)


def sigdim(dimension, depth):
    """
    Get the number of elements in the signature
    """
    if dimension == 0:
        raise ValueError("Dimension 0 is invalid")
    if depth == 1:
        return dimension
    return get_backend().sig_dim(dimension, depth)


def logsigkeys(dimension, depth):
    """
    Get the keys that correspond to the elements in the log signature
    """
    return get_backend().log_sig_keys(dimension, depth)


def sigkeys(dimension, depth):
    """
    Get the keys that correspond to the elements in the signature
    """
    return get_backend().sig_keys(dimension, depth)



import warnings

from ._tosig import *

try:
    from esig.tosig import recombine

    NO_RECOMBINE = False
except ImportError:
    NO_RECOMBINE = True

    def recombine(*args, **kwargs):
        raise NotImplementedError

warnings.warn("Importing functions from \"esig.tosig\" is deprecated and will be"
              " removed in version 1.0. Import the functions from the main \"esig\" "
              "package instead.", DeprecationWarning)



import os
import platform
#from tools import install_helpers as helpers
#from setuptools import setup, find_packages, Extension
#from setuptools.command.build_ext import build_ext
#from tools.switch_generator import SwitchGenerator

from skbuild import setup
from pathlib import Path
from setuptools import find_packages

#configuration = helpers.InstallationConfiguration(os.path.dirname(os.path.realpath(__file__)))


# SWITCH_GEN = SwitchGenerator()

#SWITCH_GEN = SwitchGenerator({
#    2: 5,
#    3: 5,
#    4: 2,
#    5: 2
#})


# class BuildExtensionCommand(build_ext):
#     """
#     Extends the build_ext class, allowing for the injection of additional code into the build_ext process.
#     """
#     def run(self):
#         """
#         Attempts to import numpy and append the result of numpy.get_include() to the self.include_dirs list.
#         This is to avoid the circular issue of importing numpy at the top of the module.
#         See https://stackoverflow.com/a/42163080.
#
#         Args:
#             self (NumpyExtensionCommand): Instance of self.
#
#         Returns:
#             None
#         """
#         print("Running extra esig pre-build commands...")
#         print("Building switch.h")
#         SWITCH_GEN.write_file()
#         print("Done")
#
#         import numpy
#         self.include_dirs.append(numpy.get_include())
#         build_ext.run(self)
#
#
#
# # https://stackoverflow.com/questions/2584595/building-a-python-module-and-linking-it-against-a-macosx-framework
# if not configuration.no_recombine and configuration.platform == helpers.PLATFORM.MACOS:
#     home = os.environ["HOME"]
#     os.environ['LDFLAGS'] = \
#         '-F ' + home + '/lyonstech/ ' + \
#         '-framework recombine ' + \
#         '-Wl,-rpath,' + home + '/lyonstech/'
#
# esig_sources = [
#     'src/tosig_module.cpp',
#     'src/Cpp_ToSig.cpp',
#     'src/ToSig.cpp',
# ]
#
# esig_depends = [
#     'src/ToSig.h',
#     'src/ToSig.cpp',
#     'src/switch.h',
# ]
#
# if not configuration.no_recombine:
#     esig_sources.extend([
#         'recombine/_recombine.cpp',
#         'recombine/TestVec/RdToPowers2.cpp'
#     ])
#     esig_depends.extend([
#         'recombine/TestVec/OstreamContainerOverloads.h',
#         'recombine/_recombine.h'
#     ])
#
# esig_extension = Extension(
#     'esig.tosig',
#     sources=esig_sources,
#     language="c++",
#     # relationship between depends and include_dirs is unclear
#     depends=esig_depends,
#     include_dirs=configuration.include_dirs,
#     library_dirs=configuration.library_dirs,
#     define_macros=configuration.define_macros,
#     libraries=configuration.used_libraries,
#     extra_compile_args=configuration.extra_compile_args,
# )
#     extra_link_args=configuration.linker_args,
#
# package_data = {
# 	"esig": ["VERSION", "ERROR_MESSAGE"]
# }
#
extras_require = {
     "iisignature": ["iisignature"],
}

# eager_resources = []
#
# if not configuration.no_recombine and configuration.platform == helpers.PLATFORM.WINDOWS:
#     package_data["esig"] += [
#         os.path.join("libiomp5md.dll"),
#         os.path.join("recombine.dll")
#     ]
#     # not sure why this is needed, and if it is, why package_data also needs to mention them
#     eager_resources += ["libiomp5md.dll", "recombine.dll"]


import io


PATH = os.path.dirname(os.path.abspath(__file__))

with io.open(os.path.join(PATH, "src", "esig", "VERSION"), "rt") as fp:
    VERSION = fp.read()

with io.open(os.path.join(PATH, "README.md"), "rt") as fp:
    DESCRIPTION = fp.read()


with io.open(os.path.join(PATH, "CHANGELOG"), "rt") as fp:
    DESCRIPTION += "\n\n\n##Changelog\n" + fp.read()



CMAKE_SETTINGS = [
    "-DLIBALGEBRA_NO_SERIALIZATION:BOOL=ON",
]
if platform.system() == "Windows":
    vcpkg = Path("build", "vcpkg")
    if vcpkg.exists():
        CMAKE_SETTINGS.append("-DCMAKE_TOOLCHAIN_FILE=%s" % (vcpkg.resolve() / "scripts" / "buildsystems" / "vcpkg.cmake"))
    elif "VCPKG_INSTALLATION_ROOT" in os.environ:
        vcpkg = Path(os.environ["VCPKG_INSTALLATION_ROOT"])
        if vcpkg.exists():
            CMAKE_SETTINGS.append("-DCMAKE_TOOLCHAIN_FILE=%s" % (vcpkg.resolve() / "scripts" / "buildsystems" / "vcpkg.cmake"))

    if platform.system() == 'Windows':
        CMAKE_SETTINGS.append("-DRECOMBINE_INSTALL_DEPENDENTS:BOOL=ON")


def filter_cmake_manifests(cmake_manifest):

    def _filter(item):
        item = str(item)
        if item.endswith(".pc"):
            return False
        elif item.endswith(".cmake"):
            return False
        elif item.endswith(".cpp"):
            return False
        elif item.endswith(".h"):
            return False
        return True

    return list(filter(_filter, cmake_manifest))



setup(
    name='esig',
    version=VERSION,

    author='Terry Lyons',
    author_email='software@lyonstech.net',
    url='http://esig.readthedocs.io/en/latest/',
    license='GPLv3',

    keywords='data streams rough paths signatures',

    description="This package provides \"rough path\" tools for analysing vector time series.",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",

    include_package_data=True,
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    test_suite='esig.tests.get_suite',

    # package_data=package_data,
    # eager_resources=eager_resources,
    # distclass=helpers.BinaryDistribution,
    # ext_modules=[esig_extension],
    cmake_process_manifest_hook=filter_cmake_manifests,

    cmake_args=CMAKE_SETTINGS,
    install_requires=['numpy>=1.7'],
    extras_require=extras_require,

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Mathematics',
        ],

    # cmdclass={
    #     'build_ext': BuildExtensionCommand,
    # }
)



import os


class SwitchGenerator:
    _path = "src"
    _default_spec = {
    #letters_range : max_depth
        (2, 2) : 20,
        (3, 3) : 10,
        (4, 4) : 6,
        (5, 6) : 6,
        (7, 9) : 5,
        (10, 16) : 4,
        (17, 25) : 2,
        #(41, 128) : 2
    }
    indent_str = "    "
    endln = "\n"
    template_fn_macro = "TemplatedFn"

    @property
    def default_spec(self):
        return dict(
            (i, d) for (mina, maxa), d in self._default_spec.items()
            for i in range(mina, maxa+1)
        )


    def __init__(self, spec=None, types=None, path=None):
        self.spec = spec or self.default_spec
        self.types = types or ["DPReal", "SPReal"]
        self.indentation = 0
        self._file = None
        self.path = path or self._path


    def _write_file(self):
        self.enter_switch("width")
        for k, v in self.spec.items():
            self.write_case(k)
            self.write_depth_switch(k, v)
        self.write_width_default()
        self.exit_switch()

    def write_depth_switch(self, w, max_depth):
        self.enter_switch("depth")
        for d in range(2, max_depth+1):
            self.write_case(d)
            self.write_templatefn(w, d, None)
            #self.write_type_switch(w, d)
        self.write_depth_default(w, max_depth)
        self.exit_switch()
        self.write_break()

    def write_type_switch(self, w, d):
        self.enter_switch("coeff")
        for typ in self.types:
            self.write_case(typ)
            self.write_templatefn(w, d, typ)
        self.write_type_default()
        self.exit_switch()

    def write_file(self):
        path = os.path.join(self.path, "switch.h")
        with open(path, "wt", encoding="UTF-8") as f:
            self._file = f
            self._write_file()
        self._file = None
        # self.write_config_bounds_header()

    def enter_switch(self, var):
        self.writeln("switch ({}) {{".format(var))
        self.indentation += 1

    def exit_switch(self):
        self.indentation -= 1
        self.writeln("}")

    def write_case(self, n):
        self.writeln("case {} :".format(n))

    def write_templatefn(self, w, d, dtype):
        text = "return {}({}, {});".format(
            self.template_fn_macro, w, d,# dtype
        )
        self.writeln(text)
        self.write_break()

    def write_break(self):
        self.writeln("break;" + self.endln)

    def write_depth_default(self, w, d):
        self.writeln("default :")
        text = "Legitimate depth of 2<->{} for records with width {} exceeds limit"
        self.writeln(
            "throw std::runtime_error ( \"{}\" );".format(
                text.format(d, w)
            )
        )

    def write_type_default(self):
        self.writeln("default :")
        text = "This type is not supported."
        self.writeln("throw std::runtime_error ( \"{}\" );".format(text))

    def write_width_default(self):
        self.writeln("default :")
        text = "Legitimate width 2 <-> 256 exceeded"
        self.writeln(
            "throw std::runtime_error ( \"{}\" );".format(text)
        )

    def writeln(self, val):
        assert self._file
        self._file.write(
            self.indent_str*self.indentation + val + self.endln
        )


    def write_config_bounds_header(self):
        path = os.path.join(self.path, "config_bounds.h")
        with open(path, "wt", encoding="UTF-8") as f:
            self._file = f
            self._write_config_bounds_header()
            self.write_checker_switch_function()
            self.write_get_bounds_switch_function()
        self._file = None

    def _write_config_bounds_header(self):
        self.start_internal_namespace()
        self.write_struct()
        for depth in self.spec:
            self.write_struct(depth)
        self.end_internal_namespace()

    def start_internal_namespace(self):
        self.writeln("namespace {")
        self.indentation += 1

    def end_internal_namespace(self):
        self.indentation -= 1
        self.writeln("}")
        self.writeln(self.endln)

    def write_struct(self, width=None):
        self.write_struct_head(width)
        self.write_struct_values(width)
        self.write_struct_end()


    def write_struct_head(self, width):
        template_str = "" if width else "DEG W"
        template_args = "<{width}>".format(width=width) if width else ""
        self.writeln("template <{template_str}>".format(template_str=template_str))
        self.writeln("struct config_check{template_args}".format(template_args=template_args))
        self.writeln("{")
        self.indentation += 1

    def write_struct_values(self, width):
        if not width:
            self.writeln("")
            return None
        min_depth = 2
        max_depth = self.spec[width]
        self.writeln("static const DEG min_depth = {min_depth};".format(min_depth=min_depth))
        self.writeln("static const DEG max_depth = {max_depth};".format(max_depth=max_depth))

    def write_struct_end(self):
        self.indentation -= 1
        self.writeln("};")
        self.writeln(self.endln)

    def write_checker_switch_function(self):
        self.writeln("inline bool check_depth_config(DEG width, DEG depth)")
        self.writeln("{")
        self.indentation += 1

        self.writeln("DEG min_depth, max_depth;")
        self.enter_switch("width")
        for width in self.spec:
            typename = "config_check<{width}>".format(width=width)
            self.write_case(width)
            self.writeln("min_depth = {typename}::min_depth;".format(typename=typename))
            self.writeln("max_depth = {typename}::max_depth;".format(typename=typename))
            self.write_break()
        self.writeln("default:")
        self.writeln("return false;")
        self.exit_switch()

        self.writeln("return (depth <= max_depth && depth >= min_depth);")
        self.indentation -= 1
        self.writeln("}")
        self.writeln(self.endln)

    def write_get_bounds_switch_function(self):
        self.writeln("inline std::pair<DEG, DEG> get_depth_bounds_config(DEG width, DEG depth)")
        self.writeln("{")
        self.indentation += 1

        self.writeln("DEG min_depth, max_depth;")
        self.enter_switch("width")
        for width in self.spec:
            typename = "config_check<{width}>".format(width=width)
            self.write_case(width)
            self.writeln("min_depth = {typename}::min_depth;".format(typename=typename))
            self.writeln("max_depth = {typename}::max_depth;".format(typename=typename))
            self.write_break()
        self.writeln("default:")
        self.writeln("return std::make_pair(0, 0);")
        self.exit_switch()

        self.writeln("return std::make_pair(min_depth, max_depth);")
        self.indentation -= 1
        self.writeln("}")
        self.writeln(self.endln)


if __name__ == "__main__":
    import sys
    g = SwitchGenerator(path=sys.argv[1])
    g.write_file()
